function [ MinRisk_labels, betas, noiseparam1, noiseparam2, currisks, ...
    WhyStop ]    = optMAC_DA(data, seed)

%{
 [ MinRisk_labels, betas, noiseparam1, noiseparam2, currisks, ...
    WhyStop ]    = optMAC_DA(data, seed)
    
    This function optimizes the Multi-Assignment Clustering (MAC) model
    with a mixture noise component. The optimizer uses deterministic
    annealing ( aka tempered EM ).
    
    input:
        - data: a NobjxNfeat binary data matrix
        - seed: optional parameter; a real number that initializes the 
                random number generator
    
    output:
        - MinRisk_labels: a (Nobj x 1) vector with the indices of the
                assignment sets yielding minimal negative log-likelihood 
                for each of the Nobj objects (likelihood factorizes)
        - betas: a (Nclusters x Nfeat) real matrix. an entry of beta 
                indicates the probability that a particular feature is not 
                member of a particular cluster centroid
        - noiseparam1: a scalar value in [0,1] indicating the global 
                probability of a bit in data to be noise. This parameter
                is called epsilon in paper.
        - noiseparam2: a scalar value in [0,1] indicating the probability
                of a noisy bit to be 1 (called r in the paper)
        - currisks = a (Niter x 1) vector logging the total negative 
                log-likelihood of each annealing step
        - WhyStop = a structure with information why annealing terminated
                (could be usefull for in-depth experiments)
%}
    
    
%% check inputs
global PRM_MAC

cont = true;

[ Nobj, Nfeat ] = size(data);
[ NassSets, Nsrc ] = size(PRM_MAC.poss_assignments);
if Nobj < Nsrc
    disp('minrisk_cluster_DA: WARNING, more possible classes than samples!');
end

if exist('seed', 'var')
    rand('seed', seed);
end



%% initialization

% set initial noise values
noiseparam1 = PRM_MAC.DetAnn.start_eps;
noiseparam2 = PRM_MAC.DetAnn.start_r;

% choose a random inizialization for the class means
indices = randperm(Nobj);
betas = 0.89*ones(Nsrc, Nfeat) - 0.78*double(data(indices(1:Nsrc),:));
% We must ensure that betas do not coincide
betas = betas + 0.10*rand(Nsrc, Nfeat);

% random initialization of the assignments
resps = 1/NassSets * ones(Nobj, NassSets);



if strcmp(PRM_MAC.DetAnn.Tmax, 'adapt')
    % use the estimate of the critical temperature, as computed in K.Rose:
    % "Deterministic annealing for Clustering, Compression, Classification,
    % Regression, and Related Optimization Problems", Proc. IEEE, 1998
    T = PRM_MAC.DetAnn.TmaxFac*sqrt(norm(single(data)));
else
    T = PRM_MAC.DetAnn.Tmax;
end

if strcmp(PRM_MAC.DetAnn.Tmin, 'adapt')
    Tmin = T / 1e10;
else
    Tmin = PRM_MAC.DetAnn.Tmin;
end

%this flag will increase cooling temperature if true and decrease it if
%false
if strcmp(PRM_MAC.DetAnn.Tfac, 'adapt')
    coolingSpeed = 'fast';
end
Tspeedlog = [];

if PRM_MAC.trace.showT
    disp([ 'starting temperature: ' num2str(T) ]);
end


%% Prepare the data structures
currisks = [];
respChangeHist = [];
noiseparam1s = [];
noiseparam2s = [];
change_resps = [];
change_betas = [];
all_resps = [];
Tvals = [];
MstepIts = [];
iterations = 0;

%% Main loop
while cont
    iterations = iterations + 1;
    old_resps = resps;
    
    
    %% check whether the centroids are very similar
    % ..and modify them (if the corresponding option is chosen)
    if PRM_MAC.DetAnn.avoidBetaCollapse && Nsrc>1
        pwDist = zeros(Nsrc);
        betaBlow = repmat(betas(1,:), [ Nsrc 1 ]);
        pwDist(:, 1) = sum(abs(betaBlow - betas), 2) ;
        for s = 2 : Nsrc
            betaBlow = repmat(betas(s,:), [ Nsrc 1 ]);
            pwDist(:, s) = sum(abs(betaBlow - betas), 2);
        end
        % place high values at diagonal and divide by Nfeat
        pwDist = pwDist / Nfeat + eye(Nsrc);
        % pwDist = (pwDist + diag( max(max(pwDist)) * ones(1,Nsrc)  ))/Nfeat;
        % find centroid wich is closest and compare with threshold
        simInds = find( min(pwDist) < 0.01 );
        
        [ rows, cols ] = ind2sub([ Nsrc, Nsrc ], simInds);
        rows = unique(rows);
        
        betas(rows,:) = betas(rows,:) ...
            + 0.1*rand(length(rows), Nfeat) - 0.05;
        
        betas(rows,:) = max(betas(rows,:), 0);
        betas(rows,:) = min(betas(rows,:), 1);
    end
    
    
    %% E-step
    [ resps, Econt ] = Estep(data, betas, T, noiseparam1, noiseparam2);
    
    if any(isnan(resps(:))) || any(isinf(resps(:)))
        Econt = false;
        disp('minrisk: Stopped because of NAN/Inf s in Estep ')
        cont = false;
        continue;
    end
    
    
    if PRM_MAC.DetAnn.update_eps && PRM_MAC.trace.DAplotResps
        figure(5);
        imagesc(resps, [0 1]);
        title('responsibilities');
        colormap gray
        colorbar
    end
    
    %% M-step
    [ betas, noiseparam1, noiseparam2, it ] = ...
        Mstep_mix(data, betas, resps, T, noiseparam1, noiseparam2);
    
    % prevent epsilon = 1;
    noiseparam1 = min(0.999, noiseparam1);
    
    if it > PRM_MAC.DetAnn.TfacItThresh
        coolingSpeed = 'slow';
    else
        coolingSpeed = 'fast';
    end
    
    
    %% compute the new risk and log the current neg. log. likelihood
    nrisk = Risk(data, betas, noiseparam1, noiseparam2);
    currisks = [ currisks; nrisk ];
    
    
    %% decrease temperature
    if strcmp(PRM_MAC.DetAnn.Tfac, 'adapt')
        if strcmp(coolingSpeed, 'fast')
            TcoolFac = PRM_MAC.DetAnn.TfacFast;
        elseif strcmp(coolingSpeed, 'slow')
            TcoolFac = PRM_MAC.DetAnn.TfacSlow;
        end
        T = T * TcoolFac;
        if PRM_MAC.trace.logData
            Tspeedlog = [ Tspeedlog TcoolFac ];
        end
    else
        T = T * PRM_MAC.DetAnn.Tfac;
    end
    
    %% monitoring
    if PRM_MAC.trace.showT
        disp([ 'New Temperature: ' num2str(T, 2) ...
            ', mean(noiseA): ' num2str(mean(noiseparam1(:)), 2) ...
            ', mean(noiseB): ' num2str(mean(noiseparam2(:)), 2) ]);
    end
    
    if PRM_MAC.trace.DAillustrate
        classAss = zeros(Nobj, Nsrc);
        [ MaxVal, MaxInd ] = max(resps, [], 2);
        for i = 1 : Nobj
            classAss(i, :) = PRM_MAC.poss_assignments(MaxInd(i), :);
        end
        if PRM_MAC.trace.DAillustrateRoundBeta
            visualize_X2ZYfactorization(data, classAss, round(1-betas), 'sorted', 1);
            figure(22); imagesc(betas); colormap('gray');
            set(gca, 'YTick', 1:size(betas,1))
            title('current centroids')
        else
            visualize_X2ZYfactorization(data, classAss, 1-betas, 'NotSorted',1 );
        end
        
    end
    
    
    %% continue ?    
    cont = Econt && (T >= Tmin) ;
    
    
end

%log stopping reason
WhyStop.Estep =   Econt;
WhyStop.Temp = (T >= Tmin);


%% Compute final risk and assignments
[ nrisk, MinRisk_labels ] = Risk(data, betas, noiseparam1, noiseparam2);
currisks = [ currisks; nrisk ];


end
